from typing import Union
import pytest
from pycobertura.utils import get_line_status

class FakeLine:
    def __init__(self, hits: str, condition_coverage: Union[str, None]):
        self.condition_coverage = condition_coverage
        self.hits = hits

    def get(self, attr: str):
        if attr == 'condition-coverage':
            return self.condition_coverage

        if attr == 'hits':
            return self.hits

        raise ValueError(f"FakeLine.get() received an unexpected attribute: {attr}")

@pytest.mark.parametrize("line, expected_output", [
    (FakeLine("0", None), "miss"),
    (FakeLine("1", None), "hit"),
    (FakeLine("1", "50% (1/2)"), "partial"),
    (FakeLine("1", "100% (2/2)"), "hit"),
    (FakeLine("0", "0% (0/2)"), "miss"),
])
def test_get_line_status(line, expected_output):
    assert get_line_status(line) == expected_output



from typing import Union
import pytest
from pycobertura.utils import calculate_line_rate, get_line_status


@pytest.mark.parametrize("total_statments, total_missing, expected_output", [
    (0, 0, 1),
    (100, 0, 1),
    (100, 1, 0.99),
    (100, 100, 0),
])
def test_calculate_line_rate(total_statments, total_missing, expected_output):
    assert calculate_line_rate(total_statments, total_missing) == expected_output



"""
Jinja2 filters meant to be used by templates.
"""

line_status_style = {
    "hit": "hit",
    "miss": "miss",
    "partial": "partial",
    None: "noop",
}


line_reason_html_icon = {
    "line-edit": "+",
    "cov-up": "&nbsp;",
    "cov-down": "&nbsp;",
    None: "&nbsp;",
}


def is_not_equal_to_dash(arg):
    return not (arg == "-")


def misses_color(arg):
    if isinstance(arg, str):
        if arg.startswith("+") or arg[0].isdigit():
            return "red"
        return "green"

    *_, status = arg
    if status == "partial":
        return "yellow"

    if status == "miss":
        return "red"

    if status == "hit":
        return "green"


def line_status(line):
    return line_status_style[line.status]


def line_reason_icon(line):
    if line.status is None:
        return "&nbsp;"
    return line_reason_html_icon[line.reason]



def test_extrapolate_coverage():
    from pycobertura.utils import extrapolate_coverage
    lines_w_status = [
        (1, "hit"),
        (4, "hit"),
        (7, "miss"),
        (9, "miss"),
    ]

    assert extrapolate_coverage(lines_w_status) == [
        (1, "hit"),
        (2, "hit"),
        (3, "hit"),
        (4, "hit"),
        (5, None),
        (6, None),
        (7, "miss"),
        (8, "miss"),
        (9, "miss"),
    ]



import mock
import runpy


def test_cli():
    with mock.patch("click.CommandCollection") as mock_cli:
        runpy.run_module("pycobertura", run_name="__main__")

    mock_cli.assert_called_once()
    mock_cli.return_value.assert_called_once_with()



import mock
import pytest
import lxml.etree as ET

from .utils import make_cobertura


def test_parse_path():
    from pycobertura import Cobertura

    xml_path = 'tests/cobertura.xml'
    with mock.patch('pycobertura.cobertura.ET.parse') as mock_parse:
        cobertura = Cobertura(xml_path)

    assert cobertura.xml is mock_parse.return_value.getroot.return_value


def test_parse_file_object():
    from pycobertura import Cobertura

    xml_path = 'tests/cobertura.xml'
    with mock.patch('pycobertura.cobertura.ET.parse') as mock_parse:
        cobertura = Cobertura(open(xml_path))

    assert cobertura.xml is mock_parse.return_value.getroot.return_value


def test_parse_string():
    from pycobertura import Cobertura

    xml_path = 'tests/cobertura.xml'
    with open(xml_path) as f:
        xml_string = f.read()
    assert ET.tostring(Cobertura(xml_path).xml) == ET.tostring(Cobertura(xml_string).xml)


def test_invalid_coverage_report():
    from pycobertura import Cobertura

    xml_path = 'non-existent.xml'
    pytest.raises(Cobertura.InvalidCoverageReport, Cobertura, xml_path)


def test_version():
    cobertura = make_cobertura()
    assert cobertura.version == '1.9'


def test_line_rate():
    cobertura = make_cobertura()
    assert cobertura.line_rate() == 0.9


def test_line_rate_by_class_file():
    cobertura = make_cobertura()
    expected_line_rates = {
        'Main.java': 1.0,
        'search/BinarySearch.java': 0.9166666666666666,
        'search/ISortedArraySearch.java': 1.0,
        'search/LinearSearch.java': 0.7142857142857143,
    }

    for filename in cobertura.files():
        assert cobertura.line_rate(filename) == \
            expected_line_rates[filename]


def test_branch_rate():
    cobertura = make_cobertura()
    assert cobertura.branch_rate() == 0.75


def test_no_branch_rate():
    from pycobertura import Cobertura

    assert Cobertura('tests/cobertura-no-branch-rate.xml').branch_rate() == None


def test_branch_rate_by_class_file():
    cobertura = make_cobertura()
    expected_branch_rates = {
        'Main.java': 1.0,
        'search/BinarySearch.java': 0.8333333333333334,
        'search/ISortedArraySearch.java': 1.0,
        'search/LinearSearch.java': 0.6666666666666666,
    }

    for filename in cobertura.files():
        assert cobertura.branch_rate(filename) == \
            expected_branch_rates[filename]


def test_missed_statements_by_class_file():
    cobertura = make_cobertura()
    expected_missed_statements = {
        'Main.java': [],
        'search/BinarySearch.java': [23, 24],
        'search/ISortedArraySearch.java': [],
        'search/LinearSearch.java': [13, 17, 19, 24],
    }

    for filename in cobertura.files():
        assert cobertura.missed_statements(filename) == \
            expected_missed_statements[filename]


def test_list_packages():
    cobertura = make_cobertura()

    packages = cobertura.packages()
    assert packages == ['', 'search']


@pytest.mark.parametrize("report, expected", [
    ('tests/cobertura-generated-by-istanbul-from-coffeescript.xml', [
        'app.coffee'
    ]),
    ('tests/cobertura.xml', [
        'Main.java',
        'search/BinarySearch.java',
        'search/ISortedArraySearch.java',
        'search/LinearSearch.java'
    ])
])
def test_list_classes(report, expected):
    cobertura = make_cobertura(xml=report)

    classes = cobertura.files()
    assert classes == expected


def test_hit_lines__by_iterating_over_classes():
    cobertura = make_cobertura()

    expected_lines = {
        'Main.java': [10, 16, 17, 18, 19, 23, 25, 26, 28, 29, 30, 31, 32, 33, 34],
        'search/BinarySearch.java': [12, 16, 18, 20, 21, 25, 26, 28, 29, 31],
        'search/ISortedArraySearch.java': [],
        'search/LinearSearch.java': [9, 15, 16],
    }

    for filename in cobertura.files():
        assert cobertura.hit_statements(filename) == expected_lines[filename]


def test_missed_lines():
    cobertura = make_cobertura()

    expected_lines = {
        'Main.java': [],
        'search/BinarySearch.java': [(23, 'partial'), (24, 'miss')],
        'search/ISortedArraySearch.java': [],
        'search/LinearSearch.java': [(13, "partial"), (17, "partial"), (19, "miss"), (20, "miss"), (21, "miss"), (22, "miss"), (23, "miss"), (24, "miss")],
    }

    for filename in cobertura.files():
        assert cobertura.missed_lines(filename) == expected_lines[filename]


def test_total_statements():
    cobertura = make_cobertura()
    assert cobertura.total_statements() == 34


def test_total_statements_by_class_file():
    cobertura = make_cobertura()
    expected_total_statements = {
        'Main.java': 15,
        'search/BinarySearch.java': 12,
        'search/ISortedArraySearch.java': 0,
        'search/LinearSearch.java': 7,
    }
    for filename in cobertura.files():
        assert cobertura.total_statements(filename) == \
            expected_total_statements[filename]


def test_total_misses():
    cobertura = make_cobertura()
    assert cobertura.total_misses() == 6


def test_total_misses_by_class_file():
    cobertura = make_cobertura()
    expected_total_misses = {
        'Main.java': 0,
        'search/BinarySearch.java': 2,
        'search/ISortedArraySearch.java': 0,
        'search/LinearSearch.java': 4,
    }
    for filename in cobertura.files():
        assert cobertura.total_misses(filename) == \
            expected_total_misses[filename]


def test_total_hits():
    cobertura = make_cobertura()
    assert cobertura.total_hits() == 28


def test_total_hits_by_class_file():
    cobertura = make_cobertura()
    expected_total_misses = {
        'Main.java': 15,
        'search/BinarySearch.java': 10,
        'search/ISortedArraySearch.java': 0,
        'search/LinearSearch.java': 3,
    }
    for filename in cobertura.files():
        assert cobertura.total_hits(filename) == \
            expected_total_misses[filename]


def test_class_file_source__sources_not_found():
    from pycobertura.cobertura import Line
    cobertura = make_cobertura()
    expected_sources = {
        'Main.java': [Line(0, 'tests/Main.java not found', None, None)],
        'search/BinarySearch.java': [Line(0, 'tests/search/BinarySearch.java not found', None, None)],
        'search/ISortedArraySearch.java': [Line(0, 'tests/search/ISortedArraySearch.java not found', None, None)],
        'search/LinearSearch.java': [Line(0, 'tests/search/LinearSearch.java not found', None, None)],
    }
    for filename in cobertura.files():
        assert cobertura.file_source(filename) == expected_sources[filename]


@pytest.mark.parametrize(
    "report, expected_line_statuses",
    [
        (
            "tests/dummy.source1/coverage.xml",
            {
                "dummy/__init__.py": [],
                "dummy/dummy.py": [
                    (1, "hit"),
                    (2, "hit"),
                    (4, "hit"),
                    (5, "miss"),
                    (6, "miss"),
                ],
                "dummy/dummy2.py": [
                    (1, "hit"),
                    (2, "hit"),
                ],
                "dummy/dummy4.py": [
                    (1, "miss"),
                    (2, "miss"),
                    (4, "miss"),
                    (5, "miss"),
                    (6, "miss"),
                ],
            },
        ),
        (
            "tests/cobertura.xml",
            {
                "Main.java": [
                    (10, "hit"),
                    (16, "hit"),
                    (17, "hit"),
                    (18, "hit"),
                    (19, "hit"),
                    (23, "hit"),
                    (25, "hit"),
                    (26, "hit"),
                    (28, "hit"),
                    (29, "hit"),
                    (30, "hit"),
                    (31, "hit"),
                    (32, "hit"),
                    (33, "hit"),
                    (34, "hit"),
                ],
                "search/BinarySearch.java": [
                    (12, "hit"),
                    (16, "hit"),
                    (18, "hit"),
                    (20, "hit"),
                    (21, "hit"),
                    (23, "partial"),
                    (24, "miss"),
                    (25, "hit"),
                    (26, "hit"),
                    (28, "hit"),
                    (29, "hit"),
                    (31, "hit"),
                ],
                "search/ISortedArraySearch.java": [],
                "search/LinearSearch.java": [(9, 'hit'), (13, 'partial'), (15, 'hit'), (16, 'hit'), (17, 'partial'), (19, 'miss'), (24, 'miss')],
            },
        ),
        (
            "tests/dummy.with-branch-condition/coverage.xml",
            {
                "__init__.py": [],
                "dummy.py": [(1, 'hit'), (2, 'partial'), (3, 'hit'), (5, 'miss')],
            },
        ),
    ],
)
def test_line_statuses(report, expected_line_statuses):
    cobertura = make_cobertura(report)
    for filename in cobertura.files():
        assert cobertura.line_statuses(filename) == \
            expected_line_statuses[filename]


@pytest.mark.parametrize("report, source, source_prefix", [
    ("tests/dummy.source1/coverage.xml", None, None),
    ("tests/dummy.source1/coverage.xml", "tests/", "dummy.source1/"),
])
def test_class_file_source__sources_found(report, source, source_prefix):
    from pycobertura.cobertura import Line
    cobertura = make_cobertura(report, source=source, source_prefix=source_prefix)
    expected_sources = {
        "dummy/__init__.py": [],
        "dummy/dummy.py": [
            Line(1, "def foo():\n", "hit", None),
            Line(2, "    pass\n", "hit", None),
            Line(3, "\n", None, None),
            Line(4, "def bar():\n", "hit", None),
            Line(5, "    a = 'a'\n", "miss", None),
            Line(6, "    b = 'b'\n", "miss", None),
        ],
        "dummy/dummy2.py": [
            Line(1, "def baz():\n", "hit", None),
            Line(2, "    pass\n", "hit", None),
        ],
        "dummy/dummy4.py": [
            Line(1, "def barbaz():\n", "miss", None),
            Line(2, "    pass\n", "miss", None),
            Line(3, "\n", None, None),
            Line(4, "def foobarbaz():\n", "miss", None),
            Line(5, "    a = 1 + 3\n", "miss", None),
            Line(6, "    pass\n", "miss", None),
        ],
    }
    for filename in cobertura.files():
        assert cobertura.file_source(filename) == \
               expected_sources[filename]


def test_class_file_source__raises_when_no_filesystem():
    from pycobertura.cobertura import Cobertura
    cobertura = Cobertura('tests/cobertura.xml')
    for filename in cobertura.files():
        pytest.raises(
            Cobertura.MissingFileSystem,
            cobertura.file_source,
            filename
        )

def test_class_source_lines__raises_when_no_filesystem():
    from pycobertura.cobertura import Cobertura
    cobertura = Cobertura('tests/cobertura.xml')
    for filename in cobertura.files():
        pytest.raises(
            Cobertura.MissingFileSystem,
            cobertura.source_lines,
            filename
        )



import pytest


@pytest.mark.parametrize(
    "line_statuses, expected_output",
    [
        ([(1, "hit")], [(1, 1, "hit")]),
        ([(1, "hit"), (2, "hit")], [(1, 2, "hit")]),
        ([(1, "hit"), (2, "hit"), (3, "hit")], [(1, 3, "hit")]),
        ([(1, "hit"), (3, "hit")], [(1, 1, "hit"), (3, 3, "hit")]),
        ([(1, "hit"), (2, "hit"), (3, "miss"), (4, "hit")], [(1, 2, "hit"), (3, 3, "miss"), (4, 4, "hit")]),
        ([(1, "hit"), (2, "partial"), (3, "miss"), (4, "hit")], [(1, 1, "hit"), (2, 2, "partial"), (3, 3, "miss"), (4, 4, "hit")]),
    ],
)
def test_rangify_by_status(line_statuses, expected_output):
    from pycobertura.utils import rangify_by_status

    assert rangify_by_status(line_statuses) == expected_output



from jinja2 import Environment, PackageLoader
from pycobertura.cobertura import Cobertura, CoberturaDiff
from pycobertura.utils import (
    green,
    rangify_by_status,
    red,
    stringify,
    calculate_line_rate,
)
from pycobertura.templates import filters
from tabulate import tabulate
from ruamel import yaml
import json
import io


env = Environment(loader=PackageLoader("pycobertura", "templates"))
env.filters["line_status"] = filters.line_status
env.filters["line_reason"] = filters.line_reason_icon
env.filters["is_not_equal_to_dash"] = filters.is_not_equal_to_dash
env.filters["misses_color"] = filters.misses_color

headers_with_missing = ["Filename", "Stmts", "Miss", "Cover", "Missing"]
headers_without_missing = ["Filename", "Stmts", "Miss", "Cover"]


class Reporter:
    def __init__(self, cobertura, ignore_regex=None):
        self.cobertura: Cobertura = cobertura
        self.ignore_regex = ignore_regex

    @staticmethod
    def format_line_rates(summary_lines):
        for i, line_rate in enumerate(summary_lines["Cover"]):
            summary_lines["Cover"][i] = f"{line_rate:.2%}"

    @staticmethod
    def format_missing_lines(summary_lines):
        for i, missing_lines in enumerate(summary_lines["Missing"]):
            summary_lines["Missing"][i] = stringify(missing_lines)

    def get_summary_lines(self):
        filenames = self.cobertura.files(ignore_regex=self.ignore_regex)
        summary_lines = {
            "Filename": filenames.copy(),
            "Stmts": [],
            "Miss": [],
            "Cover": [],
            "Missing": [],
        }
        for filename in filenames:
            file_statements = self.cobertura.total_statements(
                filename, ignore_regex=self.ignore_regex
            )
            file_misses = self.cobertura.total_misses(
                filename, ignore_regex=self.ignore_regex
            )
            file_rate = calculate_line_rate(file_statements, file_misses)
            summary_lines["Stmts"].append(file_statements)
            summary_lines["Miss"].append(file_misses)
            summary_lines["Cover"].append(file_rate)
            summary_lines["Missing"].append(self.cobertura.missed_lines(filename))

        # Generate TOTAL row
        total_statements = self.cobertura.total_statements(
            ignore_regex=self.ignore_regex
        )
        total_misses = self.cobertura.total_misses(ignore_regex=self.ignore_regex)
        total_rate = calculate_line_rate(total_statements, total_misses)
        summary_lines["Filename"].append("TOTAL")
        summary_lines["Stmts"].append(total_statements)
        summary_lines["Miss"].append(total_misses)
        summary_lines["Cover"].append(total_rate)
        summary_lines["Missing"].append([])

        return summary_lines

    def per_file_stats(self, summary_lines):
        """
        Returns dict with keys `files` and `total` that contain coverage
        statistics per file and overall, respectively.

        {
            "files": [
                {
                    "Filename": "dummy/__init__.py",
                    "Stmts": 0,
                    "Miss": 0,
                    "Cover": "0.00%",
                    "Missing": "9"
                },
                ...
            ],
            "total": {
                "Filename": "TOTAL",
                "Stmts": "...",
                "Miss": "...",
                "Cover": "..."
            }
        }
        """
        file_stats_dict_items = summary_lines.items()
        number_of_files = len(self.cobertura.files()) + 1
        file_stats_list = [
            {
                header_name: header_value[file_index]
                for header_name, header_value in file_stats_dict_items
            }
            for file_index in range(number_of_files)
        ]
        file_stats_list[-1].pop("Missing", 'Key "Missing" not found')
        return {
            "files": file_stats_list[:-1],
            "total": file_stats_list[-1],
        }


class TextReporter(Reporter):
    def generate(self):
        summary_lines = self.get_summary_lines()
        self.format_line_rates(summary_lines)
        self.format_missing_lines(summary_lines)
        return tabulate(summary_lines, headers=headers_with_missing)


class CsvReporter(Reporter):
    def generate(self, delimiter):
        summary_lines = self.get_summary_lines()
        self.format_line_rates(summary_lines)
        self.format_missing_lines(summary_lines)

        list_of_lines = [headers_with_missing]
        list_of_lines.extend(
            [[f"{item}" for item in row] for row in zip(*summary_lines.values())]
        )

        # Explanation here:
        # https://stackoverflow.com/a/55889036/9698518
        delimiter = delimiter.encode().decode("unicode_escape")

        return "\n".join([delimiter.join(line) for line in list_of_lines])


class MarkdownReporter(Reporter):
    def generate(self):
        summary_lines = self.get_summary_lines()
        self.format_line_rates(summary_lines)
        self.format_missing_lines(summary_lines)
        return tabulate(summary_lines, headers=headers_with_missing, tablefmt="github")


class JsonReporter(Reporter):
    def generate(self):
        summary_lines = self.get_summary_lines()
        self.format_line_rates(summary_lines)
        self.format_missing_lines(summary_lines)
        stats_dict = self.per_file_stats(summary_lines)
        return json.dumps(stats_dict, indent=4)


class YamlReporter(Reporter):
    def generate(self):
        summary_lines = self.get_summary_lines()
        self.format_line_rates(summary_lines)
        self.format_missing_lines(summary_lines)
        stats_dict = self.per_file_stats(summary_lines)
        # need to write to a buffer as yml packages are using a streaming interface
        buf = io.BytesIO()
        yaml.YAML().dump(stats_dict, buf)
        return buf.getvalue()


class HtmlReporter(Reporter):
    def __init__(self, *args, **kwargs):
        self.title = kwargs.pop("title", "pycobertura report")
        self.render_file_sources = kwargs.pop("render_file_sources", True)
        self.no_file_sources_message = kwargs.pop(
            "no_file_sources_message", "Rendering of source files was disabled."
        )
        super(HtmlReporter, self).__init__(*args, **kwargs)

    def generate(self):
        summary_lines = self.get_summary_lines()
        self.format_line_rates(summary_lines)
        self.format_missing_lines(summary_lines)

        filenames = summary_lines["Filename"]
        sources = []
        if self.render_file_sources:
            for i, filename in enumerate(filenames):
                if i != len(filenames) - 1:  # exclude TOTAL, not a filename
                    sources.append((filename, self.cobertura.file_source(filename)))

        template = env.get_template("html.jinja2")
        rows = {k: v[:-1] for k, v in summary_lines.items()}
        footer = {k: v[-1] for k, v in summary_lines.items()}

        return template.render(
            title=self.title,
            lines=rows,
            footer=footer,
            sources=sources,
            no_file_sources_message=self.no_file_sources_message,
        )


class DeltaReporter:
    def __init__(
        self,
        cobertura1,
        cobertura2,
        ignore_regex=None,
        show_source=True,
        *args,
        **kwargs,
    ):
        self.differ = CoberturaDiff(cobertura1, cobertura2)
        self.show_source = show_source
        self.color = kwargs.pop("color", False)
        self.ignore_regex = ignore_regex

    def format_line_rate(self, line_rate):
        return f"{line_rate:+.2%}" if line_rate else "+100.00%"

    def format_total_statements(self, total_statements):
        return f"{total_statements:+d}" if total_statements else "0"

    @staticmethod
    def format_missed_lines(missed_lines):
        output = []
        if not missed_lines:
            return output

        for lineno, status in missed_lines:
            prefix = "~" if status == "partial" else ""
            output.append(f"{prefix}{lineno:d}")
        return output

    @staticmethod
    def determine_ANSI_color_code_function_of_number(number):
        if number.startswith("+") or number[0].isdigit():
            return red
        return green

    @classmethod
    def convert_signed_number_to_ANSI_color_coded_signed_number(cls, number):
        color_function = cls.determine_ANSI_color_code_function_of_number(number)
        return color_function(number)

    def color_number(self, numbers):
        if numbers and self.color and isinstance(numbers, str):
            return self.convert_signed_number_to_ANSI_color_coded_signed_number(numbers)
        if numbers and self.color and not isinstance(numbers, str):
            return ", ".join(
                [
                    self.convert_signed_number_to_ANSI_color_coded_signed_number(number)
                    for number in numbers
                ]
            )
        return numbers if isinstance(numbers, str) else ", ".join(numbers)

    def format_total_misses(self, total_misses):
        return self.color_number(f"{total_misses:+d}") if total_misses else "0"

    def get_summary_lines(self):
        filenames = self.differ.files(ignore_regex=self.ignore_regex)
        diff_total_stmts = []
        diff_total_miss = []
        diff_total_cover = []

        for filename in filenames:
            diff_total_stmts.append(self.differ.diff_total_statements(filename))
            diff_total_miss.append(self.differ.diff_total_misses(filename))
            diff_total_cover.append(self.differ.diff_line_rate(filename))

        indexes_of_files_with_changes = [
            i
            for i in range(len(filenames))
            if any(
                (
                    diff_total_stmts[i],
                    diff_total_miss[i],
                    diff_total_cover[i],
                )
            )
        ]

        summary_lines = {
            "Filename": [],
            "Stmts": [],
            "Miss": [],
            "Cover": [],
        }

        for i in indexes_of_files_with_changes:
            summary_lines["Filename"].append(filenames[i])
            summary_lines["Stmts"].append(
                self.format_total_statements(diff_total_stmts[i])
            )
            summary_lines["Miss"].append(self.format_total_misses(diff_total_miss[i]))
            summary_lines["Cover"].append(self.format_line_rate(diff_total_cover[i]))

        summary_lines["Filename"].append("TOTAL")
        summary_lines["Stmts"].append(
            self.format_total_statements(self.differ.diff_total_statements())
        )
        summary_lines["Miss"].append(
            self.format_total_misses(self.differ.diff_total_misses())
        )
        summary_lines["Cover"].append(
            self.format_line_rate(self.differ.diff_line_rate())
        )

        if self.show_source:
            diff_total_missing = [
                self.differ.diff_missed_lines(filename) for filename in filenames
            ]
            summary_lines["Missing"] = [
                diff_total_missing[i] for i in indexes_of_files_with_changes
            ]
            summary_lines["Missing"].append("")  # for total line

        return summary_lines

    def per_file_stats(self, summary_lines):
        file_stats_dict_items = summary_lines.items()
        number_of_files = len(summary_lines["Filename"])
        file_stats_list = [
            {
                header_name: header_value[file_index]
                for header_name, header_value in file_stats_dict_items
            }
            for file_index in range(number_of_files)
        ]
        file_stats_list[-1].pop("Missing", 'Key "Missing" not found')
        return {
            "files": file_stats_list[:-1],
            "total": file_stats_list[-1],
        }


class TextReporterDelta(DeltaReporter):
    def generate(self):
        summary_lines = self.get_summary_lines()
        headers = headers_without_missing

        if self.show_source:
            missed_lines_colored = [
                self.color_number([str(m[0]) for m in missing])
                for missing in summary_lines["Missing"]
            ]

            summary_lines["Missing"] = missed_lines_colored
            headers = headers_with_missing
        return tabulate(summary_lines, headers=headers)


class CsvReporterDelta(DeltaReporter):
    def generate(self, delimiter):
        summary_lines = self.get_summary_lines()

        # lines_values: List of lines dictionary values arranged in
        # tuples of Table row values
        lines_values = list(zip(*summary_lines.values()))

        # Stringify every item in Table row values without using the Missing column
        # and store in the list list_of_lines
        list_of_lines = [headers_without_missing]
        list_of_lines.extend([[f"{item}" for item in row[:-1]] for row in lines_values])

        if self.show_source:
            # Add the Missing header to list_of_lines first inner list
            # This is a direct assignment to avoid appending an additional "Missing"
            # header in every iteration of the tests which would fail them
            list_of_lines[0] = headers_with_missing
            # Add to every list inside the list_of_lines the Missing column value
            for line_index, missing_line in enumerate(summary_lines["Missing"]):
                # for colors, explanation see here:
                # https://stackoverflow.com/a/61273717/9698518
                colored = [
                    self.color_number(str(lineno_status[0]))
                    for lineno_status in missing_line
                ]
                list_of_lines[line_index + 1] += [
                    f"{colored}".encode("utf-8").decode("unicode_escape")
                ]

        # Explanation here:
        # https://stackoverflow.com/a/55889036/9698518
        delimiter = delimiter.encode().decode("unicode_escape")

        return "\n".join([delimiter.join(line) for line in list_of_lines])


class MarkdownReporterDelta(DeltaReporter):
    def generate(self):
        summary_lines = self.get_summary_lines()
        headers = headers_without_missing

        if self.show_source:
            missed_lines_colored = [
                self.color_number([str(m[0]) for m in missing])
                for missing in summary_lines["Missing"]
            ]
            summary_lines["Missing"] = missed_lines_colored
            headers = headers_with_missing
        return tabulate(summary_lines, headers=headers, tablefmt="github")


class JsonReporterDelta(DeltaReporter):
    def generate(self):
        summary_lines = self.get_summary_lines()

        if self.show_source:
            missed_lines_colored = [
                self.color_number([str(m[0]) for m in missing])
                for missing in summary_lines["Missing"]
            ]
            summary_lines["Missing"] = missed_lines_colored

        stats_dict = self.per_file_stats(summary_lines)

        json_string = json.dumps(stats_dict, indent=4)

        # for colors, explanation see here:
        # https://stackoverflow.com/a/61273717/9698518
        colored_json_string = json_string.encode("utf-8").decode("unicode_escape")

        return colored_json_string


class YamlReporterDelta(DeltaReporter):
    def generate(self):
        summary_lines = self.get_summary_lines()

        if self.show_source:
            missed_lines_colored = [
                self.color_number([str(m[0]) for m in missing])
                for missing in summary_lines["Missing"]
            ]
            summary_lines["Missing"] = missed_lines_colored

        stats_dict = self.per_file_stats(summary_lines)
        # need to write to a buffer as yml packages are using a streaming interface
        buf = io.BytesIO()
        yaml.YAML().dump(stats_dict, buf)
        # need to replace \e escape sequence with \x1b,
        # because only the latter is supported, see also
        # the Python docs for supported formats:
        # https://docs.python.org/3/reference/lexical_analysis.html#string-and-bytes-literals
        yaml_string = buf.getvalue().replace(rb"\e", b"\x1b")
        return yaml_string


class HtmlReporterDelta(DeltaReporter):
    def __init__(self, *args, **kwargs):
        """
        Takes the same arguments as `TextReporterDelta` but also takes the keyword
        argument `show_missing` which can be set to True or False to set whether
        or not the generated report should contain a listing of missing lines in
        the summary table.
        """
        self.show_missing = kwargs.pop("show_missing", True)
        super(HtmlReporterDelta, self).__init__(*args, **kwargs)

    def generate(self):
        summary_lines = self.get_summary_lines()
        template = env.get_template("html-delta.jinja2")

        rows = {k: v[:-1] for k, v in summary_lines.items()}
        footer = {k: v[-1] for k, v in summary_lines.items()}

        render_kwargs = {
            "lines": rows,
            "footer": footer,
            "show_missing": self.show_missing,
            "show_source": self.show_source,
        }

        if self.show_source is True:
            render_kwargs["sources"] = []
            for filename in self.differ.files():
                differ_file_source_hunks = self.differ.file_source_hunks(filename)
                if differ_file_source_hunks:
                    render_kwargs["sources"].append(
                        (filename, differ_file_source_hunks)
                    )

        return template.render(**render_kwargs)


class GitHubAnnotationReporter(Reporter):
    def generate(
        self,
        annotation_level: str,
        annotation_title: str,
        annotation_message: str,
    ):
        file_names = self.cobertura.files(ignore_regex=self.ignore_regex)
        result_strs = []
        for file_name in file_names:
            for range_start, range_end, status in rangify_by_status(
                self.cobertura.missed_lines(file_name)
            ):
                result_strs.append(
                    self.to_github_annotation_message(
                        file_name=file_name,
                        start_line_num=range_start,
                        end_line_num=range_end,
                        annotation_level=annotation_level,
                        annotation_title=annotation_title,
                        annotation_message=f"{annotation_message} ({status})",
                    )
                )
        result = "\n".join(result_strs)
        return result

    @staticmethod
    def to_github_annotation_message(
        file_name: str,
        start_line_num: int,
        end_line_num: int,
        annotation_level: str,
        annotation_title: str,
        annotation_message: str,
    ):
        return f"::{annotation_level} file={file_name},line={start_line_num},endLine={end_line_num},title={annotation_title}::{annotation_message}"  # noqa


class GitHubAnnotationReporterDelta(DeltaReporter):
    def generate(
        self,
        annotation_level: str,
        annotation_title: str,
        annotation_message: str,
    ):
        summary_lines = self.get_summary_lines()
        stats_dict = self.per_file_stats(summary_lines)
        result_strs = []
        # import pdb; pdb.set_trace()

        for file_stat in stats_dict["files"]:
            for range_start, range_end, status in rangify_by_status(
                file_stat["Missing"]
            ):
                result_strs.append(
                    GitHubAnnotationReporter.to_github_annotation_message(
                        file_name=file_stat["Filename"],
                        start_line_num=range_start,
                        end_line_num=range_end,
                        annotation_level=annotation_level,
                        annotation_title=annotation_title,
                        annotation_message=f"{annotation_message} ({status})",
                    )
                )
        result = "\n".join(result_strs)
        return result
